#include "contiki.h"
#include "dev/leds.h"
#include "watchdog.h"
#include "net/rime/rime.h"

#include <stdio.h> 
#include <stdbool.h>
#include "isr_compat.h"

#include "my2c.h" // To support software I2C

#include "testbed.h" // Definition of the configuration struct for binary patching

#define LENGTH 255 // Maximum message length

/*---------------------------------------------------------------------------*/

// In flash configuration struct, will be replaced by binary patching
volatile config_t __attribute((section (".testbedConfigSection"))) cfg; // ={{{1,{208,0,0,0,0,0,0,0},{216,0,0,0,0,0,0,0},4,0,0,0,500,30000}}};

/*---------------------------------------------------------------------------*/

PROCESS(hello_world_process, "Hello world process");
AUTOSTART_PROCESSES(&hello_world_process);

/*---------------------------------------------------------------------------*/

// Interrupt service routine 
static process_event_t FALLING_EDGE_EV;

ISR(PORT2, __eeprom_isr)
{
  printf("ISR\n");
  P2IFG&=~BV(EVENT_PIN);
  process_post(&hello_world_process,FALLING_EDGE_EV,NULL);  
}

/*---------------------------------------------------------------------------*/
static char ubuffer[128]={0};
static void
recv_uc(struct unicast_conn *c, const linkaddr_t *from)
{
  static int j=0;
  int i;
  //read the payload
  packetbuf_copyto(ubuffer);
  //print the payload
  printf("unicast message received from %d.%d\n",
	 from->u8[0], from->u8[1]);
  for(j=0;j<cfg.p[0].msg_length;j++)
  	printf("%x ", (ubuffer[j]&0xff) );
  printf("\n");

  //raise gpio pin to indicate write operation
  P2OUT |= BV(EVENT_PIN);
  
  //start i2c communication
  my2c_start();
  //write address on the bus
  my2c_write(0x50<<1);
  //write memory address (2 bytes)
  my2c_write(cfg.p[0].msg_offsetH);
  my2c_write(cfg.p[0].msg_offsetL);

  //write messages up to the length in the struct
  for(i=0;i<cfg.p[0].msg_length;i++){
      my2c_write(ubuffer[i]);
  }
  //stop i2c communication
  my2c_stop();

  //lower gpio pin to indicate finished write
  P2OUT &= ~BV(EVENT_PIN);
}
static const struct unicast_callbacks unicast_callbacks = {recv_uc};
static struct unicast_conn uc;
/*---------------------------------------------------------------------------*/

PROCESS_THREAD(hello_world_process, ev, data)
{
  static struct etimer timer;
  static int i=0;
  static volatile char buffer[LENGTH]={0};
  static linkaddr_t source_addr;
  static linkaddr_t sink_addr;

  PROCESS_BEGIN();

  //allocate events for rising and falling edge
  FALLING_EDGE_EV=process_alloc_event();

  //for demo purposes, not required
  print_testbed_config((config_t*)&cfg);

  //only one pattern is used here, we copy the sink addr
  sink_addr.u8[0] = cfg.p[0].destination_id[0];
  sink_addr.u8[1] = 0;

  //only one pattern is used here, we copy the source addr
  source_addr.u8[0] = cfg.p[0].source_id[0];
  source_addr.u8[1] = 0;

  unicast_open(&uc, 146, &unicast_callbacks);

  //the software i2c may be stuck for a long time
  watchdog_stop();

  //enable and stop the i2c (otherwise it would block the bus)
  my2c_enable();
  my2c_stop();

  //node is source node
  if(linkaddr_cmp(&source_addr, &linkaddr_node_addr)) {

    //configure pin to input with interrupt
    P2DIR&=~BV(EVENT_PIN);
    P2SEL&=~BV(EVENT_PIN);
    //P2IES&=~BV(EVENT_PIN);
    P2IES|= BV(EVENT_PIN);
    P2IFG&= ~BV(EVENT_PIN);
    P2IE |= BV(EVENT_PIN);
  
    //wait until the pin is initially settled
    clock_delay(10000);
    while( (P2IN & BIT6) != 0);
    while(1) {
      printf("waiting\n");
      PROCESS_WAIT_EVENT_UNTIL(ev==FALLING_EDGE_EV);
      printf("reading %d bytes\n",(uint8_t)cfg.p[0].msg_length);
      //i2c start
      my2c_start();
      //write address on the bus
      my2c_write(0x50<<1);
      //write memory address (2 bytes)
      my2c_write(cfg.p[0].msg_offsetH);
      my2c_write(cfg.p[0].msg_offsetL);
      //disable the bus and wait a bit
      my2c_stop(); 
      clock_delay(100);
      my2c_start();
      //write address on the bus and set read bit
      my2c_write((0x50<<1)|1);
      //read back all the data, on the last byte indicate end
      for(i=0;i<(cfg.p[0].msg_length);i++){
        if(i==((cfg.p[0].msg_length)-1))
  	  buffer[i]=my2c_read(false);
  	else
          buffer[i]=my2c_read(true);
      }
      //i2c stop
      my2c_stop();

      //send the read message
      packetbuf_copyfrom((char*)buffer, (uint8_t) cfg.p[0].msg_length);
      printf("sending...\n"); 
      unicast_send(&uc, &sink_addr);
      etimer_set(&timer,CLOCK_SECOND*1);
      PROCESS_WAIT_EVENT_UNTIL(etimer_expired(&timer));

      //print the read message
      for(i=0;i<(cfg.p[0].msg_length);i++)
  	    printf("%x ", (buffer[i]&0xff) );
  

    }
  //node is sink node
  } else if(linkaddr_cmp(&sink_addr, &linkaddr_node_addr)) {
    //configure pin to output
    P2SEL &= ~BV(EVENT_PIN);
    P2DIR |=  BV(EVENT_PIN);
    P2OUT &= ~BV(EVENT_PIN);

    while(1) {
  	printf("waiting for event\n");
  	PROCESS_WAIT_EVENT();
    }
  }
  //node has no role
  else {
  while(1) {
  	printf("not playing\n");
  	PROCESS_WAIT_EVENT();
    }
  }

  PROCESS_END();
}
/*---------------------------------------------------------------------------*/
